# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Generator

from .backend import ArrayLike
from .matrixdecomposition import MatrixDecomposition, MatrixDecompositionResult
from .tensordecomposition import TensorDecomposition
from .supercore import SuperCore
from .sweepingstrategy import LocalRange
from .trainbase import TrainBase
from .utils import namespace_of_trains

def local_vector[T: ArrayLike, S: MatrixDecompositionResult](
        train: TrainBase[T],
        decomposition: MatrixDecomposition[T, S],
        ) -> Generator[T, tuple[LocalRange, T]]:
    gen = local_vector_gen(train, decomposition)
    next(gen)
    return gen

def local_vector_gen[T: ArrayLike, S: MatrixDecompositionResult](
        train: TrainBase[T],
        decomposition: MatrixDecomposition[T, S],
        ) -> Generator[T, tuple[LocalRange, T]]:
    xp = namespace_of_trains(train)
    supercore = SuperCore()
    decomp = TensorDecomposition(decomposition)
    try:
        while True:
            if len(supercore.shapes) == 0:
                sweep_data, data = yield xp.zeros(0)
            else:
                sweep_data, data = yield supercore.data
                view = [data.shape[0], *supercore.data.shape[1:-1], data.shape[-1]]
                supercore.data = xp.reshape(data, view)

            begin, end = sweep_data.begin, sweep_data.end
            if len(supercore.shapes) == 0:
                train.normalize(begin, end)
                for i in range(begin, end):
                    supercore.add_right(train.shape.digits[i], train.data[i])
                continue


            idx0 = train.shape.digits.index(supercore.shapes[0])
            idx1 = train.shape.digits.index(supercore.shapes[-1])

            slc = slice(idx0, min(begin, idx1+1))
            data = []
            norm = []
            for i in range(idx0, min(begin, idx1+1)):
                cres = supercore.cut_left(decomp)
                data.append(cres.data)
                norm.append(cres.norm)
                idx0 += 1
            train.set_data(slc, data, norm)

            slc = slice(max(end-1, idx0-1)+1, idx1+1)
            data = []
            norm = []
            for i in range(idx1, max(end-1, idx0-1), -1):
                cres = supercore.cut_right(decomp)
                data.append(cres.data)
                norm.append(cres.norm)
                idx1 -= 1
            data.reverse()
            norm.reverse()
            train.set_data(slc, data, norm)

            if len(supercore.shapes) == 0:
                train.normalize(begin, end)
                for i in range(begin, end):
                    supercore.add_right(train.shape.digits[i], train.data[i])
                continue

            for i in range(idx0-1, begin-1, -1):
                supercore.add_left(train.shape.digits[i], train.data[i])
            for i in range(idx1+1, end):
                supercore.add_right(train.shape.digits[i], train.data[i])

    except GeneratorExit:
        if len(supercore.shapes) == 0:
            return
        idxs, data, norm = [], [], []
        while len(supercore.shapes) > 0:
            cres = supercore.cut_left(decomp)
            idxs.append(train.shape.digits.index(cres.shape))
            data.append(cres.data)
            norm.append(cres.norm)
        train.set_data(slice(idxs[0], idxs[-1]+1), data, norm)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from dataclasses import dataclass
from typing import Sequence
from .trainshape import TrainShape

@dataclass(kw_only=True, frozen=True)
class ContainsResult:
    found: bool
    reverse: bool
    core_idxs: Sequence[int]

def contains(
        ref: TrainShape,
        other: TrainShape
        ) -> ContainsResult:
    found, idxs = _contains(ref, other)
    rev = False
    if not found:
        found, idxs = _contains(ref, other.reverse())
        rev = True
    if not found:
        return ContainsResult(found=False, reverse=False, core_idxs=[])
    return ContainsResult(found=True, reverse=rev, core_idxs=idxs)

def _contains(
        ref: TrainShape,
        other: TrainShape
        ) -> tuple[bool, Sequence[int]]:
    idx = 0
    map_idxs = []
    for i, digits in enumerate(ref.digits):
        if idx >= len(other):
            break
        flags = [digit in digits for digit in other.digits[idx]]
        if not any(flags):
            continue
        elif not all(flags):
            raise ValueError("Train cores do not align with operation space.")
        else:
            while all(flags):
                map_idxs.append(i)
                idx += 1
                if idx >= len(other):
                    break
                flags = [digit in digits for digit in other.digits[idx]]
    if idx < len(other):
        return False, map_idxs
    return True, map_idxs



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Any, Sequence, overload
from dataclasses import dataclass

from .backend import ArrayLike, namespace_of_arrays, get_index_dtype
from .domain import Domain
from .dimension import Dimension

@dataclass(frozen=True, init=False)
class UniformGrid:

    """
    Uniform spaced N-dimensional grid. The grid associates a dimension and a domain to each axis.
    The grid can be used to convert between coordinate and index representations of points in the grid.
    """

    #-------------------------------------------------------------------------
    #members & properties

    ndims: int
    dims: Sequence[Dimension]
    domains: Sequence[Domain]
    spacings: Sequence[float]

    #-------------------------------------------------------------------------
    #constructor

    
    @overload
    def __init__(self, dim: Dimension, domain: Domain, /) -> None: ...
    @overload
    def __init__(self, dims: Sequence[Dimension], domains: Sequence[Domain], /) -> None: ...
    # implementation
    def __init__(
            self,
            dims: Dimension | Sequence[Dimension],
            domains: Domain | Sequence[Domain], /
            ) -> None:
        if isinstance(dims, Dimension) and isinstance(domains, Domain):
            dims = [dims]
            domains = [domains]
        if len(dims) != len(domains): # type: ignore
            raise ValueError("Number of dimensions must match number of domains.")
        object.__setattr__(self, "ndims", len(dims))
        object.__setattr__(self, "dims", dims)
        object.__setattr__(self, "domains", domains)
        spacings = [self.domains[idx].diff / (self.dims[idx].size()-1) for idx in range(len(dims))]
        object.__setattr__(self, "spacings", spacings)

    #-------------------------------------------------------------------------
    #methods

    def to_coords[T: ArrayLike](self, idxs: T) -> T:
        """Convert from index representation to coordinate representation."""
        xp = namespace_of_arrays(idxs)
        int_type = get_index_dtype(xp)
        self._check_input_tensor(idxs, int_type)
        coords = xp.zeros(idxs.shape, device=idxs.device)
        for i, (dim, domain) in enumerate(zip(self.dims, self.domains)):
            vals = domain.diff / (dim.size()-1) * idxs[i,:] + domain.lower
            #coords = set_data(coords, (i, ...), vals)
            coords[i,...] = vals
                              
        return coords

    def to_idxs[T: ArrayLike](self, coords: T) -> T:
        """Convert from coordinate representation to index representation."""
        xp = namespace_of_arrays(coords)
        int_type = get_index_dtype(xp)
        self._check_input_tensor(coords, coords.dtype)
        idxs = xp.zeros(coords.shape,
                        dtype=int_type,
                        device=coords.device)
        for i, (dim, domain) in enumerate(zip(self.dims, self.domains)):
            vals = (coords[i,...] - domain.lower) / domain.diff * (dim.size()-1)
            vals = xp.asarray(xp.round(vals),
                              dtype=int_type,
                              device=coords.device)
            #idxs = set_data(idxs, (i, ...), vals)
            idxs[i,...] = vals
        return idxs

    def _check_input_tensor(self, input_tensor: ArrayLike, num_type: Any):
        if input_tensor.dtype != num_type:
            raise ValueError(f"Expected input tensor of type {num_type}, "\
                             f"got {input_tensor.dtype}")
        if input_tensor.shape[0] != self.ndims:
            raise ValueError(f"Expected input shape to be ({self.ndims}, ...), "\
                             f"got {input_tensor.shape}")



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

import numpy as np
from .trainsum import TrainSum

trainsum = TrainSum[np.typing.NDArray](np)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike
from .backend import namespace_of_arrays
from .matrixdecomposition import MatrixDecompositionResult

class QRDecomposition[T: ArrayLike]:
    """
    QR decomposition. Decomposes a matrix into an orthonormal matrix and an upper triangular matrix.
    """

    def right(self, mat: T) -> MatrixDecompositionResult[T]:
        """Calculate :math:`M=QR` and return :math:`Q` and :math:`R`."""
        q, r = self._qr(mat.T)
        #return r.T, q.T
        return MatrixDecompositionResult(left=r.T, right=q.T)

    def left(self, mat: T) -> MatrixDecompositionResult[T]:
        """Calculate :math:`M=LQ` and return :math:`L` and :math:`Q`."""
        q, r = self._qr(mat)
        return MatrixDecompositionResult(left=q, right=r)

    def _qr(self, mat: T) -> tuple[T, T]:
        xp = namespace_of_arrays(mat)
        if not hasattr(xp, "linalg"):
            raise NotImplementedError("Linalg extension missing on this backend, implement your own QRDecomposition!.")
        q, r = xp.linalg.qr(mat, mode='reduced')
        return q, r

    def left_shape(self, shape: tuple[int, int]) -> tuple[tuple[int, int], tuple[int, int]]:
        """Calculate the shape of the left function."""
        m, n = shape
        k = min(m, n)
        return (m, k), (k, n)

    def right_shape(self, shape: tuple[int, int]) -> tuple[tuple[int, int], tuple[int, int]]:
        """Calculate the shape of the right function."""
        m, n = shape
        k = min(m, n)
        return (m, k), (k, n)

    def __repr__(self) -> str:
        return f"QRDecomposition()"



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from types import EllipsisType

from .backend import Device, DType, ArrayLike, shape

class StorageTensor[T: ArrayLike]:

    _data: T
    _shape: list[int]
    _cut: list[slice]
    _view: T

    def __init__(self, storage: T) -> None:
        self._data = storage
        self._shape = [0]*len(storage.shape)
        self._cut = [slice(0,0)]*len(storage.shape)
        self._view = self._data[*self._cut]

    def __getitem__(self, idx: int | slice | EllipsisType | T | Sequence[int | slice | EllipsisType | T], /) -> T:
        return self._view[idx]

    def __setitem__(self, idx: int | slice | EllipsisType | T | Sequence[int | slice | EllipsisType | T], value: float | T) -> None:
        self._view[idx] = value

    @property
    def shape(self) -> Sequence[int]:
        return self._shape

    @property
    def max_shape(self) -> Sequence[int]:
        return shape(self._data)

    @property
    def dtype(self) -> DType:
        return self._data.dtype

    @property
    def device(self) -> Device:
        return self._data.device

    def add_to_dim(self, dim: int, size: int) -> None:
        size = self._shape[dim]+size
        if size > shape(self._data)[dim]:
            raise ValueError("Exceeding maximum size")
        self._shape[dim] = size
        self._cut[dim] = slice(0, size)
        self._view = self._data[*self._cut]



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence, Generator, Any
from itertools import product
import opt_einsum as oe

from .backend import get_index_dtype, namespace_of_arrays, ArrayLike, ArrayNamespace, device, size
from .trainshape import TrainShape
from .trainbase import TrainBase
from .localcontraction import LocalContraction

def trains_match(*trains: TrainBase) -> None:
    ref = trains[0]
    for train in trains[1:]:
        if not ref.shape == train.shape:
            raise ValueError("Trains do not match in shape, device or dtype")

def get_device_dtype(ops: Sequence[TrainBase]) -> tuple[Any, Any]:
    if len(ops) == 0:
        raise ValueError("No operands provided")
    device, dtype = ops[0].device, ops[0].dtype
    if not all(op.dtype == dtype for op in ops[1:]):
        raise ValueError("All operand dtypes must match")
    if not all(op.device == device for op in ops[1:]):
        raise ValueError("All operand devices must match")
    return device, dtype

def check_operand_shapes(ref_shapes: Sequence[TrainShape], shapes: Sequence[TrainShape]) -> None:
    if len(ref_shapes) != len(shapes):
        raise ValueError("Number of operand shapes do not match contraction")
    for ref_op, op in zip(ref_shapes, shapes):
        if ref_op != op:
            raise ValueError("Operand shapes do not match contraction")

def shape_map(shapes: Sequence[TrainShape | TrainBase], 
              *lcontrs: LocalContraction) -> dict[str, int]:
    smap: dict[str, int] = {}
    for lcontr in lcontrs:
        tns = lcontr.get_shapes(*shapes)
        for op, tn in zip(lcontr.operands, tns):
            smap.update({char: size for char, size in zip(str(op), tn)})
    return smap

def namespace_of_trains[T: ArrayLike](*trains: TrainBase[T]) -> ArrayNamespace[T]:
    return namespace_of_arrays(*(train.data[0] for train in trains))

def get_shapes(*operands: TrainShape | TrainBase) -> Sequence[TrainShape]:
    return [op.shape if isinstance(op, TrainBase) else op for op in operands]

#------------------------------------------------------------------------------------
#typing stuff

def block_tensor[T: ArrayLike](*tns: T) -> T:
    if not all(tn.ndim == tns[0].ndim for tn in tns):
        raise ValueError("All matrices must have the same number of dimensions")
    xp = namespace_of_arrays(*tns)
    index_dtype = get_index_dtype(xp)
    dev, dtype = device(tns[0]), tns[0].dtype

    dims = [xp.asarray(tn.shape, dtype=index_dtype) for tn in tns]
    shape = tuple(int(sum((dim[i] for dim in dims[1:]), start=dims[0][i])) for i in range(size(dims[0])))
    block = xp.zeros(shape, device=dev, dtype=dtype)

    off = xp.zeros(size(dims[0]), dtype=index_dtype)
    for dim, tn in zip(dims,tns):
        lower, upper = off, off + dim
        cut = tuple(slice(lower[i], upper[i]) for i in range(size(lower)))
        block[cut] = tn
        off += dim
    return block

def check_pos(msg: str, value: int | float):
    if value <= 0:
        raise ValueError(f"{msg} must be above zero, got {value}")

def check_non_neg(msg: str, value: int | float):
    if value < 0:
        raise ValueError(f"{msg} must be a positive, got {value}")

def symbol_generator() -> Generator[str]:
    idx = 0
    while True:
        yield oe.get_symbol(idx)
        idx += 1

def sequence_product(seq: Sequence[int]) -> Generator[Sequence[int]]:
    ranges = [range(s) for s in seq]
    for idxs in product(*ranges):
        yield idxs



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .trainshape import TrainShape
from .trainbase import TrainBase
from .backend import ArrayNamespace, ArrayLike, ArrayNamespace

def full[T: ArrayLike](xp: ArrayNamespace[T], shape: TrainShape, value: float) -> TrainBase[T]:
    data = []
    for i in range(len(shape)):
        data.append(xp.ones((1, *shape.middle(i), 1)))
    data[0] = value * data[0]
    return TrainBase(shape, data, copy_data=False)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Callable

from .backend import ArrayLike, namespace_of_arrays, ArrayNamespace
from .options import get_options, OptionType
from .trainshape import TrainShape
from .trainbase import TrainBase
from .evaluationexpression import EvaluationExpression as _EvaluationExpression

class EvaluateExpression[T: ArrayLike]:

    _expr: Callable

    def __init__(
        self,
        xp: ArrayNamespace[T],
        eq: str,
        *ops: TrainShape,
        ) -> None:
        opts = get_options(xp, OptionType.EVALUATE)
        self._expr = _EvaluationExpression(
                eq, *ops,
                chunk_size=opts.chunk_size,
                optimizer=opts.optimizer)

    def __call__(
            self,
            idxs: T,
            *ops: TrainBase[T],
            ) -> T:
        return self._expr(idxs, *ops)


def evaluate[T: ArrayLike](eq: str, idxs: T, *trains: TrainBase[T]) -> T:
    xp = namespace_of_arrays(idxs)
    shapes = [train.shape for train in trains]
    expr = EvaluateExpression(xp, eq, *shapes)
    return expr(idxs, *trains)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from dataclasses import dataclass

from .backend import ArrayLike
from .trainshape import TrainShape
from .trainbase import TrainBase
from .utils import get_shapes
from .einsumequation import EinsumEquation
from .einsumcontraction import EinsumContraction
from .evaluation import Evaluation
from .contractor import OptimizeKind

@dataclass(frozen=True, init=False)
class EvaluationExpression:

    equation: str
    optimizer: OptimizeKind
    chunk_size: int
    _exprs: list[Evaluation]
    _idxs: list[list[int]]
    _slices: list[list[int]]

    def __init__(
            self,
            equation: str,
            *operands: TrainShape | TrainBase,
            optimizer: OptimizeKind = "greedy",
            chunk_size: int = 1024) -> None:
        object.__setattr__(self, "equation", equation)
        object.__setattr__(self, "optimizer", optimizer)
        object.__setattr__(self, "chunk_size", chunk_size)
        shapes = get_shapes(*operands)
        org_eq = EinsumEquation(equation, *shapes)
        tmp_idxs, eqs = break_up_equation(org_eq)
        exprs = []
        op_idxs = []
        slices = []
        for idxs, eq in zip(tmp_idxs, eqs):
            contr = EinsumContraction(eq)
            expr = Evaluation(contr, optimizer=optimizer, chunk_size=chunk_size)
            expr.calc_expressions(*[operands[i] for i in idxs])
            idxs = [i for i in idxs if isinstance(operands[i], TrainShape)]
            slc = [org_eq.result.index(char) for char in contr.equation.result]
            slices.append(slc)
            exprs.append(expr)
            op_idxs.append(idxs)
        object.__setattr__(self, "_exprs", exprs)
        object.__setattr__(self, "_idxs", op_idxs)
        object.__setattr__(self, "_slices", slices)

    def __call__[T: ArrayLike](
            self,
            idxs: T,
            *operands: TrainBase[T]) -> T:
        res = self._exprs[0](idxs[self._slices[0], ...], *[operands[i] for i in self._idxs[0]])
        for i, (op_idxs, expr) in enumerate(zip(self._idxs[1:], self._exprs[1:]), start=1):
            res *= expr(idxs[self._slices[i],...], *[operands[i] for i in op_idxs])
        return res

def break_up_equation(eq: EinsumEquation) -> tuple[list[list[int]], Sequence[EinsumEquation]]:
    res_set = set(eq.result)
    op_set = list(eq.operands)
    op_idxs = list(range(len(op_set)))
    eqs = []
    eq_idxs = []
    while len(op_set) != 0:
        idxs = [len(op_set)-1]
        ref = op_set.pop()
        out = [ref]
        for i, op in enumerate(op_set):
            ovlp = set(ref) & set(op)
            if len(ovlp) > 0 and len(ovlp&res_set) != len(ovlp):
                op_set.pop(i)
                op_idxs.pop(i)
                idxs.append(i)
                out.append(op)
        res = "".join(char for char in eq.result if any(char in op for op in out))
        eq_str = ",".join(out) + f"->{res}"

        eqs.append(EinsumEquation(eq_str, *[eq.shapes[i] for i in idxs]))
        eq_idxs.append(idxs)
    return eq_idxs, eqs



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Self
from dataclasses import dataclass

@dataclass(kw_only=True)
class OperandString:
    left: str = ""
    right: str = ""
    middle: str = ""

    def reverse(self) -> Self:
        return type(self)(left="".join(reversed(self.right)),
                          middle="".join(reversed(self.middle)),
                          right="".join(reversed(self.left)))

    def __str__(self) -> str:
        return f"{self.left}{self.middle}{self.right}"




# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Generator, Sequence
from types import NoneType
from copy import deepcopy

from .backend import ArrayLike
from .localrange import LocalRange
from .trainshape import TrainShape
from .trainbase import TrainBase
from .einsumcontraction import EinsumContraction
from .operandstring import OperandString
from .localcontraction import LocalContraction
from .environment import Environment
from .sweepingstrategy import SweepingStrategy
from .contractor import ArrayContractor, OptimizeKind, DEFAULT_OPTIMIZER
from .contractorinput import ContractorInput
from .utils import shape_map
from .generatorcallabletype import GeneratorCallableType
from .contains import contains

class InnerGenerator:

    optimizer: OptimizeKind
    _contr: EinsumContraction
    _env: Environment
    _inp: NoneType | ContractorInput
    _target: int
    _cmap: Sequence[int]
    _exprs: dict[tuple[LocalRange, GeneratorCallableType], tuple[LocalContraction, ArrayContractor]]

    def __init__(self,
                 contr: EinsumContraction,
                 target: int,
                 optimizer: OptimizeKind = DEFAULT_OPTIMIZER) -> None:

        if contr.result_shape is not None or contr.full_result_shape is not None:
            raise ValueError("InnerGenerator requires an EinsumContraction without result.")
        elif target < 0 or target >= len(contr.operand_shapes):
            raise ValueError("Target index is out of bounds.")

        self.optimizer = deepcopy(optimizer)

        self._target = target
        self._contr = deepcopy(contr)
        ref = self._contr.operand_shapes[self._target]
        res = contains(self._contr.operation_shape, ref)
        self._cmap = res.core_idxs
        self._env = Environment(contr)
        self._exprs = {}
        self._inp = None

    def __call__[T: ArrayLike](
            self,
            *operands: TrainBase[T],
            expr: bool = False) -> Generator[T, tuple[LocalRange, GeneratorCallableType]]:
        gen = self._gen(*operands, calc_expr=expr)
        next(gen) # warm up
        return gen

    def calc_expressions(self,
                         strategy: SweepingStrategy,
                         *ops: TrainShape | TrainBase) -> None:
        self._inp = ContractorInput(*ops)
        ref = ops[self._target]
        ref = ref if isinstance(ref, TrainShape) else ref.shape

        self._exprs.clear()
        for _, lrange in strategy(ref):
            if lrange in self._exprs:
                continue
            self._expression(lrange, GeneratorCallableType.FULL, *ops)
            self._expression(lrange, GeneratorCallableType.LEFT, *ops)
            self._expression(lrange, GeneratorCallableType.RIGHT, *ops)
        self._env.calc_expressions(*ops)

    # ------------------------------------------------------------------------
    # Contraction generators

    def _gen[T: ArrayLike](
            self,
            *ops: TrainBase[T],
            calc_expr: bool = False) -> Generator[T, tuple[LocalRange, GeneratorCallableType]]:
        shapes = [op.shape for op in ops]
        if self._inp is None:
            self._inp = ContractorInput(*shapes)
        self._inp.check_operands(*ops)

        self._env.optimizer = self.optimizer
        env_gen = self._env(*ops, expr=calc_expr)
        data = None
        try:
            while True:
                lrange, gtype = yield data # type: ignore
                trange = self._transform_range(lrange)
                env_data = env_gen.send(trange)
                if calc_expr:
                    self._expression(lrange, gtype, *shapes)
                    loc, expr = self._exprs[lrange, gtype]
                    tns = loc.get_data(*ops, idx_map=self._inp.idx_map)
                else:
                    if (lrange, gtype) not in self._exprs:
                        self._expression(lrange, gtype, *shapes)
                    loc, expr = self._exprs[lrange, gtype]
                    tns = loc.get_data(*ops, idx_map=self._inp.idx_map)

                if gtype == GeneratorCallableType.FULL:
                    data = expr(env_data.left, *tns, env_data.right)
                elif gtype == GeneratorCallableType.LEFT:
                    data = expr(env_data.left, *tns)
                else:  # gtype == GeneratorCallableType.RIGHT
                    data = expr(*tns, env_data.right)

        except GeneratorExit:
            env_gen.close()

    # ------------------------------------------------------------------------
    # Expression builders

    def _expression(
            self,
            lrange: LocalRange,
            gtype: GeneratorCallableType,
            *ops: TrainShape | TrainBase
            ) -> None:
        trange = self._transform_range(lrange)

        left = self._contr[trange.begin].result.left
        right = self._contr[trange.end-1].result.right

        res_str = OperandString()
        lefts, rights, idxs = [], [], []

        tcontr = LocalContraction()
        for lcontr in self._contr[trange.begin:trange.end]:
            zipped = zip(lcontr.train_idxs, lcontr.core_idxs, lcontr.operands)
            for train_idx, core_idx, op_str in zipped:
                flag = lrange.begin <= core_idx < lrange.end
                if train_idx == self._target and flag:
                    res_str.middle += op_str.middle
                    lefts.append(op_str.left)
                    rights.append(op_str.right)
                    idxs.append(core_idx)
                else:
                    tcontr.add_operand(train_idx, core_idx, op_str)

        _, lefts = zip(*sorted(zip(idxs, lefts)))
        _, rights = zip(*sorted(zip(idxs, rights)))
        res_str.left = lefts[0]
        res_str.right = rights[-1]

        smap = shape_map(ops, *self._contr[trange.begin:trange.end])
        left_shape = [smap[char] for char in left]
        right_shape = [smap[char] for char in right]
        _, tns = tcontr.get_constants(*ops)

        if gtype == GeneratorCallableType.FULL:
            res_str.left  = lefts[0]
            res_str.right = rights[-1]
            eq = f"{left}," + ",".join(str(op) for op in tcontr.operands) + f",{right}->{res_str}"
            tns = [left_shape, *tns, right_shape]
        elif gtype == GeneratorCallableType.LEFT:
            res_str.left  = lefts[0]
            res_str.right = right[1:]
            eq = f"{left}," + ",".join(str(op) for op in tcontr.operands) + f"->{res_str}"
            tns = [left_shape, *tns]
        else:  # gtype == GeneratorCallableType.RIGHT
            res_str.left  = left[1:]
            res_str.right = rights[-1]
            eq = ",".join(str(op) for op in tcontr.operands) + f",{right}->{res_str}"
            tns = [*tns, right_shape]

        tcontr.set_result(res_str)
        contr = ArrayContractor(eq, *tns, optimizer=self.optimizer)
        self._exprs[lrange, gtype] = tcontr, contr

    def _transform_range(self, lrange: LocalRange) -> LocalRange:
        begin, end = self._cmap[lrange.begin], self._cmap[lrange.end-1]+1
        return LocalRange(begin=begin, end=end)



import unittest
from itertools import product

from trainsum import TrainSum
from utils import backends, rand_data

class TestSVDecomposition(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [(120, 120), (256, 250), (120, 7), (256, 800)]

    def test_construction(self):
        for ts in self.trainsum:
            decomp = ts.svdecomposition(max_rank=2, cutoff=1e-10)
            self.assertRaises(ValueError, ts.svdecomposition, max_rank=-1, cutoff=1e-10)
            self.assertRaises(ValueError, ts.svdecomposition, max_rank=2, cutoff=-1e-10)
    
    def test_decompose(self):
        mrank = 2
        cutoff = 1e-10
        for ts, sizes in product(self.trainsum, self.sizes):
            xp = ts.namespace
            decomp = ts.svdecomposition(max_rank=mrank, cutoff=cutoff)

            mat = xp.asarray(rand_data(xp, *sizes))
            res = decomp.left(mat)
            u, s, vh = xp.linalg.svd(mat, full_matrices=False)
            numel = min(mrank, xp.sum(s > cutoff))

            ref_left = u[:,:numel]
            ref_right = s[:numel,xp.newaxis] * vh[:numel,:]
            self.assertTrue(xp.max(res.left - ref_left) < 1e-10)
            self.assertTrue(xp.max(res.right - ref_right) < 1e-10)

            res = decomp.right(mat)
            ref_left = u[:,:numel] * s[xp.newaxis,:numel]
            ref_right = vh[:numel,:]
            self.assertTrue(xp.max(res.left - ref_left) < 1e-10)
            self.assertTrue(xp.max(res.right - ref_right) < 1e-10)

if __name__ == '__main__':
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike, shape
from .trainbase import TrainBase
from .utils import trains_match, block_tensor, sequence_product, get_device_dtype, namespace_of_trains

class ExactAddition:

    def __call__[T: ArrayLike](self,
                 *trains: TrainBase[T]
                 ) -> TrainBase[T]:
        trains_match(*trains)
        xp = namespace_of_trains(*trains)
        device, dtype = get_device_dtype(trains)

        data = []
        for idx in range(len(trains[0].shape)):
            is_begin = (idx == 0)
            is_end = (idx == len(trains[0].shape)-1)

            tns = [train.data[idx] for train in trains]
            shapes = [shape(tn) for tn in tns]

            left = 1 if is_begin else sum(sh[0] for sh in shapes)
            middle = shapes[0][1:-1]
            right = 1 if is_end else sum(sh[-1] for sh in shapes)
            tmp = xp.zeros((left, *middle, right),
                           device=device, dtype=dtype)

            left_cut = 0 if is_begin else slice(None)
            right_cut = 0 if is_end else slice(None)
            for idxs in sequence_product(middle):
                cut = (left_cut, *idxs, right_cut)
                tmp[cut] = block_tensor(*[tn[cut] for tn in tns])
            data.append(tmp)

        return TrainBase(trains[0].shape, data, copy_data=False)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Callable, Generator, Sequence
from copy import deepcopy

from .backend import ArrayLike
from .localrange import LocalRange
from .trainshape import TrainShape
from .trainbase import TrainBase
from .einsumcontraction import EinsumContraction
from .sweepingstrategy import SweepingStrategy
from .contractor import OptimizeKind, DEFAULT_OPTIMIZER
from .einsumequation import EinsumEquation
from .linearmapgenerator import LinearMapGenerator
from .generatorcallabletype import GeneratorCallableType

class LinearMap[T: ArrayLike]:

    optimizer: OptimizeKind
    _ops: Sequence[TrainBase]

    @property
    def result(self) -> TrainShape:
        return self._map_gen.result_shape

    def __init__(
            self,
            eq: str,
            *ops: TrainShape | TrainBase[T],
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER) -> None:
        self.optimizer = deepcopy(optimizer)
        self._ops, self._map_gen = self._get_map_gen(eq, *ops)

    def __call__(
            self,
            guess: TrainBase[T],
            expr: bool = False
            ) -> Generator[Callable[[T], T], tuple[LocalRange, GeneratorCallableType]]:
        return self._map_gen(*self._ops, guess, expr=expr)

    def calc_expressions(
            self,
            strat: SweepingStrategy,
            guess: TrainShape | TrainBase) -> None:
        guess_shape = guess if isinstance(guess, TrainShape) else guess.shape
        self._map_gen.calc_expressions(strat, *[op.shape for op in self._ops], guess_shape)

    def _get_map_gen(self, eq: str, *ops: TrainShape | TrainBase[T]) -> tuple[Sequence[TrainBase], LinearMapGenerator]:
        idx = self._ref_idx(*ops)
        ops_, guess = self._split_ops(idx, *ops)

        tmp = eq.split("->")
        op_strs = tmp[0].split(",")
        guess_str = op_strs.pop(idx)
        eq = ",".join(op_strs) +f",{guess_str}->{tmp[1]}"

        einsum_eq = EinsumEquation(eq, *[op.shape for op in ops_], guess)
        contr = EinsumContraction(einsum_eq)
        return ops_, LinearMapGenerator(contr, idx, optimizer=self.optimizer)

    def _split_ops(self, idx: int, *ops: TrainShape | TrainBase) -> tuple[list[TrainBase], TrainShape]:
        ops_ = list(deepcopy(op) for op in ops)
        ref = ops_.pop(idx)
        return ops_, ref # type: ignore

    def _ref_idx(self, *ops: TrainShape | TrainBase[T]) -> int:
        idx = -1
        for i, op in enumerate(ops):
            if isinstance(op, TrainShape):
                idx = i
                break
        if idx == -1 or sum(isinstance(op, TrainShape) for op in ops) != 1:
            raise ValueError("One of the operands must be a TrainShape.")
        return idx



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike
from .options import OptionType, get_options, ExactOptions, VariationalOptions
from .trainbase import TrainBase
from .utils import namespace_of_trains

from .exactaddition import ExactAddition
from .decompositionaddition import DecompositionAddition
from .variationaladdition import VariationalAddition

def add[T: ArrayLike](train1: TrainBase[T], train2: TrainBase[T], *trains: TrainBase[T]) -> TrainBase[T]:
    xp = namespace_of_trains(train1, train2, *trains)
    opts = get_options(xp, OptionType.EINSUM)
    if isinstance(opts, ExactOptions):
        return ExactAddition()(train1, train2, *trains)
    if opts.decomposition is None or opts.strategy is None:
        raise ValueError("Decomposition and strategy must be provided for 'decomposition' and 'variational' methods.")
    add = DecompositionAddition(opts.strategy, opts.decomposition)
    guess = add(train1, train2, *trains)
    if isinstance(opts, VariationalOptions):
        var_add = VariationalAddition(train1.shape, opts.decomposition, opts.strategy, opts.optimizer)
        guess = var_add(guess, train1, train2, *trains)
    return guess



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Protocol, Sequence, Callable
from .backend import ArrayLike

class LocalEigSolverResult[T: ArrayLike](Protocol):
    """Protocol for the result of a local eigenvalue solver."""

    #: The eigenvector corresponding to the computed eigenvalue.
    array: T

    #: The computed eigenvalue.
    value: float

class LocalEigSolver[T: LocalEigSolverResult](Protocol):
    """Protocol for a local eigenvalue solver."""

    def __call__[S: ArrayLike](self,
                 mat: Callable[[S], S],
                 guess: S,
                 states: Sequence[S] = [], /
                 ) -> T:
        """
        Solve an eigenvalue problem for a linear map with an initial guess and optional states to orthogonalize against.
        """
        ...



import unittest

from trainsum import TrainSum
from utils import backends

class TestDomain(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]

    def test_construction(self):
        for ts in self.trainsum:
            self.assertRaises(ValueError, ts.domain, 1.0, 1.0)
            self.assertRaises(ValueError, ts.domain, 2.0, 1.0)
            lower = 1.0
            upper = 2.0
            domain = ts.domain(lower, upper)
            self.assertEqual(domain.lower, lower)
            self.assertEqual(domain.upper, upper)
            self.assertEqual(domain.diff, upper - lower)

    def test_eq(self):
        for ts in self.trainsum:
            domain1 = ts.domain(1.0, 2.0)
            domain2 = ts.domain(1.0, 2.0)
            domain3 = ts.domain(0.0, 2.0)
            self.assertEqual(domain1, domain2)
            self.assertNotEqual(domain1, domain3)

if __name__ == '__main__':
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from __future__ import annotations
from typing import Callable, Literal
from itertools import product

from .backend import DType, Device, get_index_dtype, namespace_of_arrays, size, ArrayLike, device, shape
from .matrixleastsquares import MatrixLeastSquares
from .functionalcore import FunctionalCore
from .storagetensor import StorageTensor

Driver = Literal["gels"] | Literal["gelsd"]

class PiCore[T: ArrayLike]:
    _data: StorageTensor[T]

    _rows: tuple[StorageTensor[T], StorageTensor[T]]
    _cols: tuple[StorageTensor[T], StorageTensor[T]]

    _row_mask: StorageTensor[T]
    _col_mask: StorageTensor[T]

    _row_map: dict[tuple[int, ...], tuple[int, ...]]
    _col_map: dict[tuple[int, ...], tuple[int, ...]]

    @property
    def dtype(self) -> DType:
        return self._data.dtype

    @property
    def device(self) -> Device:
        return self._data.device

    def __init__(
            self,
            lcore: FunctionalCore[T],
            rcore: FunctionalCore[T],
            func: Callable[[T], T],
            dtype: DType) -> None:
        left, lmid, rmid, right = lcore.left, lcore.middle, rcore.middle, rcore.right

        xp = namespace_of_arrays(left.data[...])
        dev = device(left.data)
        data = xp.zeros((left.max_size, lmid.size, rmid.size, right.max_size),
                        device=dev, dtype=dtype)
        self._data = StorageTensor(data)

        self._data.add_to_dim(1, lmid.size)
        self._data.add_to_dim(2, rmid.size)

        self._row_map = {}
        self._col_map = {}

        int_type = get_index_dtype(xp)
        self._rows = (StorageTensor(xp.zeros(left.max_size*lmid.size, device=dev, dtype=int_type)),
                      StorageTensor(xp.zeros(left.max_size*lmid.size, device=dev, dtype=int_type)))
        self._cols = (StorageTensor(xp.zeros(right.max_size*rmid.size, device=dev, dtype=int_type)),
                      StorageTensor(xp.zeros(right.max_size*rmid.size, device=dev, dtype=int_type)))

        info = xp.__array_namespace_info__()
        btype = info.dtypes()["bool"]

        self._row_mask = StorageTensor(xp.zeros((left.max_size, lmid.size), device=dev, dtype=btype))
        self._row_mask.add_to_dim(1, lmid.size)
        self._col_mask = StorageTensor(xp.zeros((rmid.size, right.max_size), device=dev, dtype=btype))
        self._col_mask.add_to_dim(0, rmid.size)

        self.update_rows(lcore, rcore, func)
        self.update_cols(lcore, rcore, func)

        for i in range(lcore.right.size):
            self.set_col(lcore.right.data[:,i])
        for i in range(rcore.left.size):
            self.set_row(rcore.left.data[:,i])
    
    def update_rows(
            self,
            lcore: FunctionalCore,
            rcore: FunctionalCore,
            func: Callable[[T], T]) -> None:
        left, lmid, rmid, right = lcore.left, lcore.middle, rcore.middle, rcore.right
        added = left.size - self._data.shape[0]
        ndims = lmid.ndims
        if added == 0:
            return

        xp = namespace_of_arrays(self._data[...])
        int_type = get_index_dtype(xp)
        ax = xp.newaxis
        idxs = xp.zeros((ndims, added, lmid.size),
                        device=self.device, dtype=int_type)
        idxs += left.data[:,-added:,ax]
        idxs += lmid.data[:,ax,:]
        for i, j in product(range(added), range(lmid.size)):
            key = tuple(int(val) for val in idxs[:,i,j])
            self._row_map[key] = (i+self._data.shape[0], j)
        self._data.add_to_dim(0, added)
        self._row_mask.add_to_dim(0, added)
        if size(self._data[...]) == 0:
            return

        rsize = self._data.shape[3]
        idxs = idxs[:,:,:,ax,ax] + rmid.data[:,ax,ax,:,ax] + right.data[:,ax,ax,ax,:rsize]
        self._data[-added:,:,:,:] = func(idxs)

    def update_cols(
            self,
            lcore: FunctionalCore,
            rcore: FunctionalCore,
            func: Callable[[T], T]) -> None:
        left, lmid, rmid, right = lcore.left, lcore.middle, rcore.middle, rcore.right
        added = right.size - self._data.shape[3]
        ndims = lcore.middle.ndims
        if added == 0:
            return

        xp = namespace_of_arrays(self._data[...])
        int_type = get_index_dtype(xp)
        ax = xp.newaxis
        idxs = xp.zeros((ndims, rmid.size, added),
                        device=self.device, dtype=int_type)
        idxs += rmid.data[:,:,ax]
        idxs += right.data[:,ax,-added:]
        for i, j in product(range(rmid.size), range(added)):
            key = tuple(int(val) for val in idxs[:,i,j])
            self._col_map[key] = (i, j+self._data.shape[3])
        self._data.add_to_dim(3, added)
        self._col_mask.add_to_dim(1, added)
        if size(self._data[...]) == 0:
            return

        lsize = self._data.shape[0]
        idxs = left.data[:,:lsize,ax,ax,ax] + lmid.data[:,ax,:,ax,ax] + idxs[:,ax,ax,:,:]
        self._data[:,:,:,-added:] = func(idxs)

    def get_error(self, solver: MatrixLeastSquares) -> T:
        left = self._data[:,:,self._col_mask[...]]
        right = self.inv_right(solver)
        xp = namespace_of_arrays(left, right)
        approx = xp.tensordot(left, right, axes=([2],[0]))
        return xp.abs(approx - self._data[...])

    def set_row(self, vals: T) -> None:
        key = tuple(int(val) for val in vals)
        if key not in self._row_map:
            return
        left_idx, middle_idx = self._row_map[key]
        if self._row_mask[left_idx, middle_idx]:
            return
        self._rows[0].add_to_dim(0, 1)
        self._rows[0][-1] = left_idx
        self._rows[1].add_to_dim(0, 1)
        self._rows[1][-1] = middle_idx
        self._row_mask[left_idx, middle_idx] = True

    def set_col(self, vals: T) -> None:
        key = tuple(int(val) for val in vals)
        if key not in self._col_map:
            return
        middle_idx, right_idx = self._col_map[key]
        if self._col_mask[middle_idx, right_idx]:
            return
        self._cols[0].add_to_dim(0, 1)
        self._cols[0][-1] = middle_idx
        self._cols[1].add_to_dim(0, 1)
        self._cols[1][-1] = right_idx
        self._col_mask[middle_idx, right_idx] = True

    def inv_right(self, solver: MatrixLeastSquares) -> T:
        xp = namespace_of_arrays(self._data[...])
        cmask = xp.reshape(self._col_mask[...], (size(self._col_mask[...]),))
        clen = int(xp.sum(cmask))
        if clen == size(cmask):
            view = (clen, *self._data.shape[2:])
            return xp.reshape(xp.eye(clen, device=self.device, dtype=self.dtype), view)

        right = self.right()
        right = xp.reshape(right, (shape(right)[0], shape(right)[1]*shape(right)[2]))
        q = xp.linalg.qr(right.T)[0].T
        A = q[:,cmask]
        B = q[:,~cmask]
        lres = solver(A, B)
        
        res = xp.zeros((clen, self._data.shape[2]*self._data.shape[3]),
                       device=self.device, dtype=self.dtype)
        res[:,cmask] = xp.eye(clen, device=self.device, dtype=self.dtype)
        res[:,~cmask] = lres

        view = (clen, *self._data.shape[2:])
        return xp.reshape(res, view)

    def _solve(self, A: T, B: T) -> T:
        xp = namespace_of_arrays(A, B)
        lres = xp.linalg.lstsq(A, B) #type: ignore
        return lres[0]

    def left(self) -> T:
        return self._data[:,:,self._cols[0][...],self._cols[1][...]]

    def right(self) -> T:
        return self._data[self._rows[0][...],self._rows[1][...],:,:]

    def result_right(self, solver: MatrixLeastSquares) -> T:
        xp = namespace_of_arrays(self._data[...])
        int_type = get_index_dtype(xp)
        right = self.inv_right(solver)
        idxs = xp.zeros(self._cols[0].shape, device=self.device, dtype=int_type)
        idxs += self._cols[0][...] * shape(self._data[...])[3]
        idxs += self._cols[1][...]
        sorted_idxs = xp.sort(idxs)
        res_idxs = xp.searchsorted(sorted_idxs, idxs)
        return right[res_idxs,:,:]



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Hashable, Literal, Any, Self, overload
from enum import Enum
from copy import deepcopy
import threading

from .backend import ArrayNamespace
from .contractor import OptimizeKind, DEFAULT_OPTIMIZER
from .matrixdecomposition import MatrixDecomposition
from .sweepingstrategy import SweepingStrategy
from .matrixleastsquares import MatrixLeastSquares

class OptionType(Enum):
    EINSUM = 0
    CROSS = 1
    EVALUATE = 2

class Options:

    key: Hashable

    def __init__(self, namespace: ArrayNamespace, category: OptionType):
        self.key = (namespace, category, threading.get_ident())

    def __enter__(self) -> Self:
        global _opts
        if self.key in _opts:
            self._tmp = _opts[self.key]
        else:
            self._tmp = None
        _opts[self.key] = self
        return self

    def __exit__(self, *_) -> None:
        global _opts
        if self._tmp is not None:
            _opts[self.key] = self._tmp
        else:
            del _opts[self.key]

class ExactOptions(Options):
    """
    Context manager for exact einsum options.
    """

    #: Optimizer for the local einsum contraction paths.
    optimizer: OptimizeKind

    def __init__(
            self, *,
            namespace: ArrayNamespace,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER):
        self.optimizer = deepcopy(optimizer)
        super().__init__(namespace, OptionType.EINSUM)

class EvaluationOptions(Options):
    """
    Context manager for evalaute einsum options.
    """

    #: Optimizer for the local einsum contraction paths.
    optimizer: OptimizeKind
    #: Size of the chunks that are evaluated together.
    chunk_size: int

    def __init__(
            self, *,
            namespace: ArrayNamespace,
            chunk_size: int = 1024,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER):
        self.optimizer = deepcopy(optimizer)
        self.chunk_size = chunk_size
        super().__init__(namespace, OptionType.EVALUATE)


class DecompositionOptions[T: MatrixDecomposition](Options):
    """
    Context manager for decomposition based einsum options. The decomposition and strategy are used
    to determine the ranks and the approximation of the einsum result.
    """
    #: Decomposition method for determining the approximation of the einsum result.
    decomposition: T
    #: Sweeping strategy for determining the path of the algorithm.
    strategy: SweepingStrategy
    #: Optimizer for the local einsum contraction paths.
    optimizer: OptimizeKind

    def __init__(
            self, *,
            namespace: ArrayNamespace,
            decomposition: T,
            strategy: SweepingStrategy, 
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER):
        self.decomposition = deepcopy(decomposition)
        self.strategy = deepcopy(strategy)
        self.optimizer = deepcopy(optimizer)
        super().__init__(namespace, OptionType.EINSUM)

class VariationalOptions[T: MatrixDecomposition](DecompositionOptions):
    """
    Context manager for variational based einsum options.
    """
    #: Decomposition method for determining the approximation of the einsum result.
    decomposition: T
    #: Sweeping strategy for determining the path of the algorithm.
    strategy: SweepingStrategy
    #: Optimizer for the local einsum contraction paths.
    optimizer: OptimizeKind
    pass

class CrossOptions[T: MatrixLeastSquares](Options):
    """
    Context manager for cross approximation options.
    """
    #: Sweeping strategy for determining the path of the algorithm.
    strategy: SweepingStrategy
    #: Convergence criterion.
    eps: float
    #: Least squares solver for solving the local problems in the cross interpolation.
    solver: T
    def __init__(
            self, *,
            namespace: ArrayNamespace,
            strategy: SweepingStrategy,
            eps: float,
            solver: T):
        self.solver = deepcopy(solver)
        self.eps = deepcopy(eps)
        self.strategy = deepcopy(strategy)
        super().__init__(namespace, OptionType.CROSS)

_opts: dict[Any, Options] = {}

@overload
def get_options(namespace: ArrayNamespace, otype: Literal[OptionType.EINSUM]) -> ExactOptions | DecompositionOptions | VariationalOptions: ...
@overload
def get_options(namespace: ArrayNamespace, otype: Literal[OptionType.CROSS]) -> CrossOptions: ...
@overload
def get_options(namespace: ArrayNamespace, otype: Literal[OptionType.EVALUATE]) -> EvaluationOptions: ...
# implementation
def get_options(namespace: ArrayNamespace, otype: OptionType) -> Options:
    global _opts
    key = (namespace, otype, threading.get_ident())
    if key in _opts:
        return _opts[key]
    else:
        raise KeyError("No options set for the current thread.")

def set_options(opts: CrossOptions | EvaluationOptions | ExactOptions | DecompositionOptions | VariationalOptions) -> None:
    global _opts
    _opts[opts.key] = opts




# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Callable, Generator
from copy import deepcopy

from .backend import ArrayLike
from .localrange import LocalRange
from .trainshape import TrainShape, change_dims
from .trainbase import TrainBase
from .einsumequation import EinsumEquation
from .einsumcontraction import EinsumContraction
from .operandstring import OperandString
from .localcontraction import LocalContraction
from .environment import Environment
from .contractor import ArrayContractor, OptimizeKind, DEFAULT_OPTIMIZER
from .sweepingstrategy import SweepingStrategy
from .utils import shape_map, get_shapes, check_operand_shapes
from .contractorinput import ContractorInput
from .generatorcallabletype import GeneratorCallableType
from .contains import contains

class LinearMapGenerator[T: ArrayLike]:

    @property
    def result_shape(self) -> TrainShape:
        return self._ref_shape

    optimizer: OptimizeKind
    _contr: EinsumContraction
    _target: int
    _inp: None | ContractorInput = None
    _ref_shape: TrainShape
    _exprs: dict[tuple[LocalRange, GeneratorCallableType], tuple[LocalContraction, ArrayContractor]]

    def __init__(
            self,
            contr: EinsumContraction,
            target: int,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER) -> None:

        self.optimizer = deepcopy(optimizer)
        self._contr = self._get_contr(contr, target)
        self._env = Environment(self._contr)
        self._target = target
        self._ref_shape = self._contr.operand_shapes[0]

        res = contains(self._contr.operation_shape, self._contr.operand_shapes[0])
        self._lcmap = res.core_idxs
        res = contains(self._contr.operation_shape, self._contr.operand_shapes[-1])
        self._rcmap = res.core_idxs
        if self._lcmap != self._rcmap:
            raise NotImplementedError("Only linear maps, where the input digits reside on the "\
                                      "same cores as the output digits are currently supported.")
        self._exprs = {}

    def __call__(
            self,
            *ops: TrainBase[T],
            expr: bool = False
            ) -> Generator[Callable[[T], T], tuple[LocalRange, GeneratorCallableType]]:
        gen = self._gen(*ops, calc_expr=expr)
        next(gen)  # warm up
        return gen

    def calc_expressions(self, strat: SweepingStrategy, *ops: TrainShape | TrainBase) -> None:
        ops_ = list(ops)
        guess = ops_.pop(self._target)
        if not isinstance(guess, TrainShape):
            guess = guess.shape
        ops_ = [guess, *ops_ , guess]
        shapes = get_shapes(*ops_)
        check_operand_shapes(self._contr.operand_shapes, shapes)

        self._inp = ContractorInput(*ops_)
        self._exprs = {}
        for _, lrange in strat(guess):
            if lrange not in self._exprs:
                self._expression(lrange, GeneratorCallableType.FULL, *ops_)
                self._expression(lrange, GeneratorCallableType.LEFT, *ops_)
                self._expression(lrange, GeneratorCallableType.RIGHT, *ops_)
        self._env.calc_expressions(*ops_)
    
    
    # ------------------------------------------------------------------------
    # Contraction generators

    def _gen(
            self,
            *ops: TrainBase[T],
            calc_expr: bool = False
            ) -> Generator[Callable[[T], T], tuple[LocalRange, GeneratorCallableType]]:
        ops_ = list(ops)
        guess = ops_.pop(self._target)
        ops_ = [guess, *ops_ , guess]
        shapes = get_shapes(*ops_)
        if self._inp is None:
            self._inp = ContractorInput(*shapes)
        self._inp.check_operands(*ops_)

        env_gen = self._env(*ops_)
        func = lambda x: x
        try:
            while True:
                lrange, mtype = yield func
                if calc_expr or lrange not in self._exprs:
                    self._expression(lrange, mtype, *shapes)
                env_data = env_gen.send(lrange)
                tcontr, expr = self._exprs[lrange, mtype]
                tns = tcontr.get_data(*ops_, idx_map=self._inp.idx_map)
                if mtype == GeneratorCallableType.FULL:
                    func = lambda x: expr(env_data.left, *tns, x, env_data.right)
                elif mtype == GeneratorCallableType.LEFT:
                    func = lambda x: expr(env_data.left, *tns, x)
                else:  # mtype == GeneratorCallableType.RIGHT:
                    func = lambda x: expr(*tns, x, env_data.right)
        finally:
            env_gen.close()

    # ------------------------------------------------------------------------
    # Expression builders

    def _expression(self, lrange: LocalRange, mtype: GeneratorCallableType, *ops: TrainShape | TrainBase) -> None:
        shapes = get_shapes(*ops)
        trange = self._transform_range(lrange)

        left = self._contr[trange.begin].result.left
        right = self._contr[trange.end-1].result.right

        inp_str = OperandString(left=left[-1], right=right[-1])
        res_str = OperandString()
        tcontr = LocalContraction()
        for lcontr in self._contr[trange.begin:trange.end]:
            zipped = zip(lcontr.train_idxs, lcontr.core_idxs, lcontr.operands)
            for train_idx, core_idx, op_str in zipped:
                flag = lrange.begin <= core_idx < lrange.end
                if train_idx == 0 and flag:
                    res_str.middle += op_str.middle
                elif train_idx == len(self._contr.operand_shapes)-1 and flag:
                    inp_str.middle += op_str.middle
                else:
                    tcontr.add_operand(train_idx, core_idx, op_str)

        smap = shape_map(shapes, *self._contr[trange.begin:trange.end])
        left_shape = [smap[char] for char in left]
        right_shape = [smap[char] for char in right]
        inp_shape = [smap[char] for char in str(inp_str)]
        _, tns = tcontr.get_constants(*ops)

        if mtype == GeneratorCallableType.FULL:
            res_str.left  = left[0]
            res_str.right = right[0]
            eq = f"{left}," + ",".join(str(op) for op in tcontr.operands) + f",{inp_str},{right}->{res_str}"
            tns = [left_shape, *tns, inp_shape, right_shape]
        elif mtype == GeneratorCallableType.LEFT:
            res_str.left  = left[0]
            res_str.right = right[1:]
            eq = f"{left}," + ",".join(str(op) for op in tcontr.operands) + f",{inp_str}->{res_str}"
            tns = [left_shape, *tns, inp_shape]
        else:  # mtype == GeneratorCallableType.RIGHT
            res_str.left  = left[1:]
            res_str.right = right[0]
            eq = f",".join(str(op) for op in tcontr.operands) + f",{inp_str},{right}->{res_str}"
            tns = [*tns, inp_shape, right_shape]

        #tcontr.set_result(res_str)
        expr = ArrayContractor(eq, *tns, optimizer=self.optimizer)
        self._exprs[lrange, mtype] = tcontr, expr

    def _transform_range(self, lrange: LocalRange) -> LocalRange:
        lbegin, lend = self._lcmap[lrange.begin], self._lcmap[lrange.end-1]+1
        rbegin, rend = self._rcmap[lrange.begin], self._rcmap[lrange.end-1]+1
        begin = min(lbegin, rbegin)
        end = max(lend, rend)
        return LocalRange(begin=begin, end=end)

    def _get_contr(self, contr: EinsumContraction, target: int) -> EinsumContraction:
        ops = [*contr.operand_shapes]
        ref = ops.pop(target)
        ops = [ref, *ops, ref]
        op_strs = [str(op) for op in contr.equation.operands]
        ref_op_str = op_strs.pop(target)
        eq = ",".join([contr.equation.result, *op_strs, ref_op_str]) + "->"
        eq = EinsumEquation(eq, *ops)

        dim_map = {}
        for in_shape, out_shape in zip(ops, eq.shapes):
            for in_dim, out_dim in zip(in_shape.dims, out_shape.dims):
                dim_map[in_dim] = out_dim
        new_op_shape_dims = [dim_map[dim] for dim in contr.operation_shape.dims]
        op_shape = change_dims(contr.operation_shape, new_op_shape_dims)
        return EinsumContraction(eq, op_shape=op_shape)

    #def _get_ops(self, *ops: TrainShape)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

import sys
from typing import Protocol, overload, SupportsIndex, Iterator, Sequence, Any
from dataclasses import dataclass

class SequenceOfLike[T](Protocol):

    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[T]: ...
    def __reversed__(self) -> Iterator[T]: ...
    @overload
    def __getitem__(self, idx: SupportsIndex) -> T: ...
    @overload
    def __getitem__(self, idx: slice) -> Sequence[T]: ...
    def __contains__(self, item: Any) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def index(self, value: T, start: SupportsIndex = 0, stop: SupportsIndex = sys.maxsize) -> int: ...
    def count(self, value: T) -> int: ...

@dataclass(frozen=True, init=False)
class SequenceOf[T](Sequence):

    _seq_data: list[T]

    def __init__(self, data: Sequence[T]) -> None:
        object.__setattr__(self, "_seq_data", data)

    #-------------------------------------------------------------------------
    #container behaviour

    def __len__(self) -> int:
        return len(self._seq_data)

    def __iter__(self) -> Iterator[T]:
        return self._seq_data.__iter__()

    def __reversed__(self) -> Iterator[T]:
        return self._seq_data.__reversed__()

    @overload
    def __getitem__(self, idx: SupportsIndex) -> T: ...
    @overload
    def __getitem__(self, idx: slice) -> Sequence[T]: ...
    #implementation
    def __getitem__(self, idx: SupportsIndex | slice) -> T | Sequence[T]:
        return self._seq_data[idx]

    def __contains__(self, item: Any) -> bool:
        return len(self) > 0\
               and isinstance(item, type(self._seq_data[0]))\
               and item in self._seq_data

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self))\
               and self._seq_data == other._seq_data

    def index(self, value: T, start: SupportsIndex = 0, stop: SupportsIndex = sys.maxsize) -> int:
        return self._seq_data.index(value, start, stop)

    def count(self, value: T) -> int:
        return self._seq_data.count(value)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Self, Callable, overload, Sequence, Optional
from copy import deepcopy

from trainsum.utils import namespace_of_trains

from .backend import ArrayLike, Device, DType, ArrayNamespace, namespace_of_arrays
from .trainshape import TrainShape
from .trainbase import TrainBase
from .utils import symbol_generator

from .full import full
from .add import add
from .multiply import multiply
from .matmul import matmul
from .evaluate import evaluate
from .transform import transform
from .construct import construct
from .to_tensor import to_tensor
from .fuse import fuse
from .conj import conj
from .truncate import truncate

class TensorTrain[S: ArrayLike]:
    """
    N-dimensional tensor train. Main class for representing and manipulating tensor trains.
    Should not be instantiated directly, but rather through the `tensortrain` function.
    """

    _base: TrainBase[S]

    @property
    def device(self) -> Device:
        """Get and set the device."""
        return self._base.device
    @device.setter
    def device(self, device: Device) -> None:
        self._base.device = device

    @property
    def dtype(self) -> DType:
        """Get and set the data type."""
        return self._base.dtype
    @dtype.setter
    def dtype(self, dtype: DType) -> None:
        self._base.dtype = dtype

    @property
    def shape(self) -> TrainShape:
        """Return the shape of the tensor train with the current ranks."""
        return self._base.shape

    @property
    def cores(self) -> Sequence[S]:
        """3-dimensional tensor cores."""
        return self._base.data

    def __init__(self,
                 base: TrainBase[S],
                 copy_data: bool = True) -> None:
        self._base = deepcopy(base) if copy_data else base

    # ------------------------------------------------------------------------
    # multiplication

    def __imul__(self, other: int | float | Self, /) -> Self:
        if isinstance(other, TensorTrain):
            base = multiply(self._base, other._base)
            self._base = base
        else:
            self._base.data[0][...] *= other
        return self

    def __mul__(self, other: int | float | Self, /) -> Self:
        return deepcopy(self).__imul__(other)

    def __rmul__(self, other: int | float | Self, /) -> Self:
        return deepcopy(self).__imul__(other)

    # ------------------------------------------------------------------------
    # multiplication

    def __imatmul__(self, other: Self, /) -> Self:
        base = matmul(self._base, other._base)
        self._base = base
        return self

    def __matmul__(self, other: Self, /) -> Self:
        return deepcopy(self).__imatmul__(other)

    # ------------------------------------------------------------------------
    # addition

    def __iadd__(self, other: int | float | Self, /) -> Self:
        if isinstance(other, TensorTrain):
            self._base = add(self._base, other._base)
        else:
            xp = namespace_of_trains(self._base)
            base = full(xp, self._base.shape, other)
            self._base = add(self._base, base)
        return self

    def __add__(self, other: int | float | Self, /) -> Self:
        return deepcopy(self).__iadd__(other)

    def __radd__(self, other: int | float | Self, /) -> Self:
        return deepcopy(self).__iadd__(other)

    # ------------------------------------------------------------------------
    # divide

    def __itruediv__(self, other: int | float | Self, /) -> Self:
        other = 1.0/other
        return self.__imul__(other)

    def __truediv__(self, other: int | float | Self, /) -> Self:
        return deepcopy(self).__itruediv__(other)

    def __rtruediv__(self, other: int | float | Self, /) -> Self:
        base = transform(self._base, lambda x: 1.0/x) # type: ignore
        return type(self)(base, copy_data=False) * other

    # ------------------------------------------------------------------------
    # unary operators

    def __pos__(self) -> Self:
        return self.__mul__(1)

    def __neg__(self) -> Self:
        return self.__mul__(-1)

    # ------------------------------------------------------------------------
    # cross based

    def __ipow__(self, power: int, /) -> Self:
        if power == 2:
            return self.__imul__(self)

        self._base = transform(self._base, lambda x: x ** power)
        return self

    def __pow__(self, power: int, /) -> Self:
        return deepcopy(self).__ipow__(power)

    def __abs__(self) -> Self:
        base = transform(self._base, lambda x: abs(x))
        return type(self)(base, copy_data=False)

    # ------------------------------------------------------------------------
    # getter & setter

    def __getitem__(self, cut: tuple[S], /) -> S:
        sgen = symbol_generator()
        chars = "".join(next(sgen) for _ in range(len(self._base.shape.dims)))
        eq = f"{chars}->{chars}"
        xp = namespace_of_arrays(*cut)
        idxs = xp.stack(list(cut), axis=0)
        return evaluate(eq, idxs, self._base)

    # ------------------------------------------------------------------------
    # other

    def to_tensor(self) -> S:
        """Construct the full tensor from the tensor train."""
        return to_tensor(self._base)

    def extend(self, train: Self, /) -> None:
        """Extend the tensor train by fusing it with another tensor train."""
        self._base = fuse(self._base, train._base)

    def conj(self) -> Self:
        """Return the complex conjugate of the tensor train."""
        base = conj(self._base)
        return type(self)(base, copy_data=False)

    def normalize(self, idx: int, /) -> None:
        """Create the canonical form of the tensor train with respect to the core at index idx."""
        self._base.normalize(idx)

    def truncate(self) -> None:
        """Reduce the ranks of the tensor train according to the current einsum options."""
        self._base = truncate(self._base)

    def transform(self, func: Callable[[S], S]) -> Self:
        """Perform an element-wise transformation of the tensor train defined by some function."""
        base = transform(self._base, func)
        return type(self)(base, copy_data=False)

    def __repr__(self) -> str:
        return f"TensorTrain: {self._base.shape}"


@overload
def tensortrain[T: ArrayLike](shape: TrainShape, func: Callable[[T], T], start_idxs: T, xp: ArrayNamespace[T], /) -> TensorTrain[T]: ...
@overload
def tensortrain[T: ArrayLike](shape: TrainShape, data: Sequence[T], /) -> TensorTrain[T]: ...
@overload
def tensortrain[T: ArrayLike](shape: TrainShape, data: T, /) -> TensorTrain[T]: ...
# implementation
def tensortrain[T: ArrayLike](
        shape: TrainShape,
        data: T | Sequence[T] | Callable[[T], T],
        start_idxs: Optional[T] = None,
        xp: Optional[ArrayNamespace[T]] = None,
        /) -> TensorTrain[T]:
    if isinstance(data, Sequence):
        base = TrainBase(shape, data, copy_data=False)
    elif isinstance(data, Callable):
        if xp is None:
            raise ValueError("Array namespace must be provided when data is a function.")
        base = construct(shape, data, xp, start_idxs)
    else:
        base = construct(shape, data)
    return TensorTrain(base, copy_data=False)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike
from .trainbase import TrainBase
from .utils import symbol_generator
from .einsum import einsum

def multiply[T: ArrayLike](train1: TrainBase[T], train2: TrainBase[T]) -> TrainBase[T]:
    sgen = symbol_generator()
    chars = "".join(next(sgen) for _ in range(len(train1.shape.dims)))
    eq = f"{chars},{chars}->{chars}"
    res = einsum(eq, train1, train2)
    if not isinstance(res, TrainBase):
        raise RuntimeError("Unexpected result type from EinsumExpression.")
    return res



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Generator, Sequence, Callable, Optional
from copy import deepcopy
from string import ascii_lowercase
from math import prod

from .backend import ArrayLike, shape
from .localrange import LocalRange
from .trainbase import TrainBase
from .linearmap import LinearMap
from .locallinsolver import LocalLinSolver, LocalLinSolverResult
from .utils import get_device_dtype, namespace_of_trains

from .trainshape import TrainShape, change_dims
from .innergenerator import InnerGenerator
from .sweepingstrategy import SweepingStrategy
from .einsumequation import EinsumEquation
from .einsumcontraction import EinsumContraction
from .contractor import OptimizeKind, DEFAULT_OPTIMIZER
from .generatorcallabletype import GeneratorCallableType
from .direction import Direction
from .tensordecomposition import TensorDecomposition
from .operationspace import einsum_operation_shape
from .matrixdecomposition import MatrixDecomposition
from .normalization import Normalization
from .svdecomposition import SVDecomposition

class AMEnSolver[T: ArrayLike, S: LocalLinSolverResult]:
    
    @property
    def strategy(self) -> SweepingStrategy:
        return self._strategy
    @strategy.setter
    def strategy(self, value: SweepingStrategy) -> None:
        if value.ncores != 1:
            raise ValueError("AMEnSolver only supports single-core sweeping strategies.")
        self._strategy = deepcopy(value)

    solver: LocalLinSolver[S]
    optimizer: OptimizeKind
    decomposition: MatrixDecomposition
    _strategy: SweepingStrategy
    _maps: Sequence[LinearMap]
    _rhs_gen: InnerGenerator
    _rhs: TrainBase[T]

    def __init__(
            self,
            solver: LocalLinSolver[S],
            rhs: TrainBase[T],
            *maps: LinearMap[T],
            strategy: SweepingStrategy = SweepingStrategy(),
            decomposition: MatrixDecomposition = SVDecomposition(max_rank=2),
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER) -> None:

        self.solver = deepcopy(solver)
        self.strategy = strategy
        self.optimizer = deepcopy(optimizer)
        self.decomposition = deepcopy(decomposition)
        self._maps = maps
        self._rhs = deepcopy(rhs)
        self._rhs_gen = self._get_gen(rhs.shape)

    def __call__(
            self,
            guess: TrainBase[T],
            callback: Optional[Callable[[LocalRange, S], bool]] = None,
            ) -> TrainBase[T]:
        guess = deepcopy(guess)
        gen = self._gen(guess)
        next(gen)  # warm up
        for _, lrange in self.strategy(guess.shape):
            loc_res = gen.send(lrange)
            if callback is not None:
                if callback(lrange, loc_res):
                    break
        return guess

    def _gen(
            self,
            guess: TrainBase[T],
            ) -> Generator[S, LocalRange]:
        xp = namespace_of_trains(guess)
        device, dtype = get_device_dtype([guess])
        decomp = TensorDecomposition(self.decomposition)

        map_gens = [lin_map(guess) for lin_map in self._maps]
        rhs_gen = self._rhs_gen(guess, self._rhs)
        gen_type = GeneratorCallableType.FULL

        loc_res = None
        direc = Direction.TO_RIGHT
        idx = -1
        try:
            while True:
                lrange = yield loc_res # type: ignore
                if lrange.end - lrange.begin != 1:
                    raise ValueError("AMEn only supports single-core sweeps.")
                if lrange.begin == 0:
                    direc = Direction.TO_RIGHT
                elif lrange.end == len(guess.shape):
                    direc = Direction.TO_LEFT
                else:
                    direc = Direction.TO_LEFT if lrange.begin < idx else Direction.TO_RIGHT
                idx = lrange.begin

                guess.normalize(idx)
                funcs = [gen.send((lrange, gen_type)) for gen in map_gens]
                func = lambda x: sum((f(x) for f in funcs), start=xp.zeros_like(x))
                lrhs = rhs_gen.send(((lrange, gen_type)))
                loc_res = self.solver(func, lrhs, guess.data[idx])
                tmp = loc_res.array

                if direc == Direction.TO_RIGHT:

                    num_left = len(tmp.shape)-1
                    lops = []
                    for gen in map_gens:
                        op_data = gen.send((lrange, GeneratorCallableType.LEFT))(tmp)
                        op_data = xp.reshape(op_data, (*tmp.shape[:-1], prod(op_data.shape[num_left:])))
                        lops.append(op_data)
                    lrhs = rhs_gen.send((lrange, GeneratorCallableType.LEFT))

                    right = sum(shape(op)[-1] for op in lops) + shape(lrhs)[-1]
                    exact = xp.zeros((*tmp.shape[:-1], right), device=device, dtype=dtype)

                    start = 0
                    for op in lops:
                        end = start + prod(op.shape[num_left:])
                        exact[...,start:end] = op
                        start = end
                    exact[...,start:] = lrhs
                    u = decomp.left(exact, -1).left
                    numel = u.shape[-1]

                    lshape = guess.data[idx].shape
                    left = xp.zeros((*lshape[:-1], lshape[-1]+numel), device=device, dtype=dtype)
                    left[...,:-numel] = tmp
                    left[...,-numel:] = u

                    rshape = guess.data[idx+1].shape
                    right = xp.zeros((rshape[0]+numel, *rshape[1:]), device=device, dtype=dtype)
                    right[:-numel,:,:] = guess.data[idx+1]

                    guess.set_data(slice(idx, idx+2), [left, right], [Normalization.LEFT, Normalization.NONE])

                else:  # Direction.TO_LEFT

                    num_right = len(tmp.shape)-1
                    lops = []
                    for gen in map_gens:
                        op_data = gen.send((lrange, GeneratorCallableType.RIGHT))(tmp)
                        op_data = xp.reshape(op_data, (prod(op_data.shape[:-num_right]), *tmp.shape[1:]))
                        lops.append(op_data)
                    lrhs = rhs_gen.send((lrange, GeneratorCallableType.RIGHT))

                    left = sum(shape(op)[0] for op in lops) + shape(lrhs)[0]
                    exact = xp.zeros((left, *tmp.shape[1:]), device=device, dtype=dtype)

                    start = 0
                    for op in lops:
                        end = start + op.shape[0]
                        exact[start:end,...] = op
                        start = end
                    exact[start:,...] = lrhs
                    vh = decomp.right(exact, 1).right
                    numel = vh.shape[0]

                    rshape = guess.data[idx].shape
                    right = xp.zeros((rshape[0]+numel, *rshape[1:]), device=device, dtype=dtype)
                    right[:-numel,...] = tmp
                    right[-numel:,...] = vh

                    lshape = guess.data[idx-1].shape
                    left = xp.zeros((*lshape[:-1], lshape[-1]+numel), device=device, dtype=dtype)
                    left[...,:-numel] = guess.data[idx-1]

                    guess.set_data(slice(idx-1, idx+1), [left, right], [Normalization.NONE, Normalization.RIGHT])


        finally:
            for gen in map_gens:
                gen.close()
            rhs_gen.close()

    def _get_gen(self, state: TrainShape) -> InnerGenerator:
        dims = ascii_lowercase[:len(state.dims)]
        eq = EinsumEquation(f"{dims},{dims}->{dims}", state, state)
        op_shape = einsum_operation_shape(eq)
        eq_ = EinsumEquation(f"{dims},{dims}->", state, state)
        dim_map = {}
        for in_shape, out_shape in zip(eq.shapes, eq_.shapes):
            for in_dim, out_dim in zip(in_shape.dims, out_shape.dims):
                dim_map[in_dim] = out_dim
        new_op_shape_dims = [dim_map[dim] for dim in op_shape.dims]
        contr = EinsumContraction(eq_, op_shape=change_dims(op_shape, new_op_shape_dims))
        return InnerGenerator(contr, 0, self.optimizer)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike
from .trainbase import TrainBase
from .utils import namespace_of_trains

def conj[T: ArrayLike](train: TrainBase[T]) -> TrainBase[T]:
    xp = namespace_of_trains(train)
    data_conj = [xp.conj(d) for d in train.data]
    return TrainBase(train.shape, data_conj)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from math import comb

from .backend import ArrayNamespace, ArrayLike
from .uniformgrid import UniformGrid
from .trainshape import TrainShape
from .trainbase import TrainBase
from .full import full

def polyval[T: ArrayLike](xp: ArrayNamespace[T], grid: UniformGrid, coeffs: Sequence[float], offset: float) -> TrainBase[T]:
    if grid.ndims != 1:
        raise ValueError("Polynomial function only supports 1D uniform grid")
    dim = grid.dims[0]
    shape = TrainShape(dim, [(d,) for d in dim])

    domain = grid.domains[0]
    spacing = grid.spacings[0]
    off = (domain.lower-offset)/(len(dim))
    order = len(coeffs)

    if len(coeffs) == 1:
        return full(xp, shape, next(iter(coeffs)))
    
    cores = []
    for digit in dim:
        xi = spacing*digit.factor
        data = xp.zeros((order, digit.base, order))
        for j in range(order):
            for k in range(digit.base):
                row_data = xp.asarray(_get_row(order-j, k*xi+off))
                #data = set_data(data, (j, k, slice(j, None)), row_data)
                data[j,k,j:] = row_data
        cores.append(data)
        
    tmp = xp.asarray(coeffs).reshape(1, order)
    cores[0] = xp.tensordot(tmp, cores[0], axes=([1], [0]))
    tmp = xp.asarray([*[0.0]*(order-1), 1]).reshape(order, 1)
    idx = len(cores[-1].shape)-1
    cores[-1] = xp.tensordot(cores[-1], tmp, axes=([idx], [0]))
    return TrainBase(shape, cores, copy_data=False)

def _get_row(num: int, val: float) -> list[float]:
    row = [0.0]*num
    row[0] = 1
    for i in range(1, num):
        row[i] = comb(num-1, num-i-1) * val**i
    return row



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Any, Sequence
from abc import abstractmethod
from dataclasses import dataclass

from .backend import ArrayLike
from .dimension import Dimension

@dataclass(frozen=True, init=False)
class Grid:

    ndims: int
    dims: Sequence[Dimension]

    def index_dtype(self) -> Any: ...

    @abstractmethod
    def to_coords[T: ArrayLike](self, idxs: T) -> T: ...
    @abstractmethod
    def to_idxs[T: ArrayLike](self, coords: T) -> T: ...



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence, overload, Optional, Callable
from copy import deepcopy

from .backend  import ArrayLike, namespace_of_arrays, shape_size
from .matrixdecomposition import MatrixDecompositionResult, MatrixDecomposition

class TensorDecomposition[T: ArrayLike, ResType: MatrixDecompositionResult]:

    _mat_decomp: MatrixDecomposition[T, ResType]

    def __init__(self, decomposition: MatrixDecomposition[T, ResType]) -> None:
        self._mat_decomp = deepcopy(decomposition)

    def _decompose(self,
             decomp: Callable[[T], ResType],
             tn: T,
             left_dims: int | Sequence[int],
             right_dims: Optional[Sequence[int]] = None, /
             ) -> ResType:
        xp = namespace_of_arrays(tn)
        if isinstance(left_dims, int) and right_dims is None:
            rdims = tn.shape[left_dims:]
            ldims = tn.shape[:left_dims]
        elif not isinstance(left_dims, int) and right_dims is not None:
            tn = xp.permute_dims(tn, (*left_dims, *right_dims))
            rdims = tn.shape[len(left_dims):]
            ldims = tn.shape[:len(left_dims)]
        else:
            raise ValueError("Invalid input arguments")
        mat = xp.reshape(tn, (shape_size(ldims), shape_size(rdims)))
        res = decomp(mat)
        res.left = xp.reshape(res.left, (*ldims, res.left.shape[1]))
        res.right = xp.reshape(res.right, (res.right.shape[0], *rdims))
        return res

    @overload
    def left(self, tn: T, split: int, /) -> ResType: ...
    @overload
    def left(self, tn: T, left_dims: Sequence[int], right_dims: Sequence[int], /) -> ResType: ...
    # impl
    def left(self,
             tn: T,
             left_dims: int | Sequence[int],
             right_dims: Optional[Sequence[int]] = None, /
             ) -> ResType:
        return self._decompose(self._mat_decomp.left, tn, left_dims, right_dims)

    @overload
    def right(self, tn: T, split: int, /) -> ResType: ...
    @overload
    def right(self, tn: T, left_dims: Sequence[int], right_dims: Sequence[int], /) -> ResType: ...
    # impl
    def right(self,
              tn: T,
              left_dims: int | Sequence[int],
              right_dims: Optional[Sequence[int]] = None, /
              ) -> ResType:
        return self._decompose(self._mat_decomp.right, tn, left_dims, right_dims)

    def _shape(self,
               shape_func: Callable[[tuple[int, int]], tuple[tuple[int, int], tuple[int, int]]],
               shape: tuple[int, ...],
               left_dims: int | Sequence[int],
               right_dims: Optional[Sequence[int]] = None, /
               ) -> tuple[tuple[int, ...], tuple[int, ...]]:
        if isinstance(left_dims, int):
            rdims = shape[left_dims:]
            ldims = shape[:left_dims]
        elif not isinstance(left_dims, int) and right_dims is not None:
            ldims = tuple(shape[i] for i in left_dims)
            rdims = tuple(shape[i] for i in right_dims)
        else:
            raise ValueError("Invalid input arguments")
        left_shape, right_shape = shape_func((shape_size(ldims), shape_size(rdims)))
        return (*ldims, left_shape[-1]), (right_shape[0], *rdims)

    @overload
    def left_shape(self,
                   shape: tuple[int, ...],
                   split: int, /
                   ) -> tuple[tuple[int, ...], tuple[int, ...]]: ...
    @overload
    def left_shape(self,
                   shape: tuple[int, ...],
                   left_dims: Sequence[int],
                   right_dims: Sequence[int], /
                   ) -> tuple[tuple[int, ...], tuple[int, ...]]: ...
    # impl
    def left_shape(self,
             shape: tuple[int, ...],
             left_dims: int | Sequence[int],
             right_dims: Optional[Sequence[int]] = None, /
             ) -> tuple[tuple[int, ...], tuple[int, ...]]:
        return self._shape(self._mat_decomp.left_shape, shape, left_dims, right_dims)

    @overload
    def right_shape(self,
                    shape: tuple[int, ...],
                    split: int, /
                    ) -> tuple[tuple[int, ...], tuple[int, ...]]: ...
    @overload
    def right_shape(self,
                    shape: tuple[int, ...],
                    left_dims: Sequence[int],
                    right_dims: Sequence[int], /
                    ) -> tuple[tuple[int, ...], tuple[int, ...]]: ...
    # impl
    def right_shape(self,
             shape: tuple[int, ...],
             left_dims: int | Sequence[int],
             right_dims: Optional[Sequence[int]] = None, /
             ) -> tuple[tuple[int, ...], tuple[int, ...]]:
        return self._shape(self._mat_decomp.right_shape, shape, left_dims, right_dims)
